(*Expert mode values*)
let interrupt_on_exception = false 
let defaultExtArraySize = ref 5
let defaultGraphSize = ref 5
let defaultLiftSetSize = ref 5
let defaultHeapSize = ref 5
let defaultLogSize = ref 10
let debugModeOn = ref false
let progressBarSymbol = ref '#'
let progressBarSize = ref 60
let plotSepChar = ref ' '
let dumpIfDeadlocked = ref false
let maxConsecutiveClash = ref 10
let backtrace = ref false
let (rescale:int option ref) = ref None
let eclipseMode = ref false
let useColor = ref true
let safeModeOn = ref false
let emacsMode = ref false

(* expert option for stories *)
  (** Memory **)
  (* Number of potential states that are put in the cache per binding site, so as to handler with side effects in stories. None -> Unlimited cache *) 
  let cache_size = ref (None:int option) 

  (** Precomputation **)
  (* Cut concurrent events (for all observables) before generating the blackboard *)
  let do_global_cut = true

  (* Cut pseudo-inverse events *) 
  let cut_pseudo_inverse_event = true
    
  (* Cut concurrent events (for the current observale) before generating the blackboard *)
  let do_local_cut = true

  (* Cut separable components *)
  let do_detect_separable_components = true
   
  (** Propagation heuristics **)
  (* Whenever we do not know whether an event has to be selected or, not, check whether this is not the last one that can parform a requested action *)
  let look_up_for_better_cut = true 
  
  (* Whenever an event is removed, checked whether there is not only one left to perform a required action *)   
  let look_down_for_better_cut = true 

  let log_number_of_causal_flows = true 

(*User definable values*)
let (maxEventValue:int option ref) = ref None
let (maxTimeValue:float option ref) = ref None
let (pointNumberValue:int option ref) = ref None
let (seedValue:int option ref) = ref None
let plotModeOn = ref false
let compileModeOn = ref false
let implicitSignature = ref false
let dotOutput = ref false
let fluxModeOn = ref false
let snapshotHighres = ref true

let causalModeOn = ref false
let weakCompression = ref false 
let strongCompression = ref false 
let mazCompression = ref false 
let showIntroEvents = ref false

(*XLS output for the grids during compression*)
let dump_grid_before_weak_compression = false
let dump_grid_before_strong_compression = false
let dump_grid_after_branching_during_weak_compression = false
let dump_grid_after_branching_during_strong_compression = false
let xlsweakFileName = "grid_weak_compression"
let xlsstrongFileName = "grid_strong_compression"

(*Computed values*)
let (timeIncrementValue:float option ref) = ref None

(***)
let (syncTime:float ref) = ref 1.0
let (syncMode:string option ref)= ref None
let (showProgressStep:int option ref) = ref None
(***)

let cpuTime = ref 0.0
let initSimTime () = cpuTime := Sys.time ()  

(*Name convention*)
let outputDirName = ref ""
let snapshotFileName = ref "snap"

let dumpFileName = ref "dump.ka"
let cflowFileName = ref "cflow.dot" 
let profilingName = ref "profiling.txt" 
let with_weak_compression = "weakly_compressed"
let without_compression = "without_compression" 
let influenceFileName = ref ""
let fluxFileName = ref ""
let outputDataName = ref ""
let inputKappaFileNames:(string list ref) = ref [] 
let marshalizedInFile = ref "" 
let marshalizedOutFile = ref ""


let set name ext_opt = 
	if !name = "" then ()
	else
	let fname = Filename.concat !outputDirName !name in
	let fname = 
		match ext_opt with
			| None -> fname
			| Some ext ->
				if (Filename.check_suffix fname ext) then fname 
				else 
				(fname^"."^ext)
	in
	name:=fname

let setOutputName () = 
	set snapshotFileName (Some "dot");
	set dumpFileName (Some "ka");
	set influenceFileName (Some "dot") ;
	set fluxFileName (Some "dot") ;
	set marshalizedOutFile None;
	set cflowFileName (Some "dot") ; 
	set outputDataName None

let checkFileExists () =
	let check file =  
		match file with
			| "" -> ()
			| file -> 
				if Sys.file_exists file then 
					begin
						Printf.fprintf stderr "File '%s' already exists do you want to erase (y/N)? \n" file ; flush stderr ; 
						let answer = Tools.read_input () in
						if answer="y" then () else exit 1
					end
				else ()
	in
	check !influenceFileName ;
	check !fluxFileName ;
	check !marshalizedOutFile ; 
	let points = match !pointNumberValue with None -> false | Some _ -> true in
	if points then check !outputDataName 

let (openOutDescriptors:out_channel list ref) = ref []
let (openInDescriptors:in_channel list ref) = ref []

let add_out_desc d = openOutDescriptors := d::!openOutDescriptors  
let add_in_desc d = openInDescriptors := d::!openInDescriptors  

type current_compression_mode = Weak | Strong | Causal
type compression_mode = 
    { 
      causal_trace:bool;
      weak_compression:bool;
      strong_compression:bool
    }
      
let get_compression_mode () = 
  {
    causal_trace=(!mazCompression);
    weak_compression=(!weakCompression);
    strong_compression=(!strongCompression);
  }

let get_causal_trace x = x.causal_trace 
let get_causal_trace_only x = not (x.weak_compression || x.strong_compression)
let get_weak_compression x = x.weak_compression
let get_strong_compression x = x.strong_compression
let get_cache_size x = !cache_size 
